function [B, C, time, err] = panda(A, k, r, v)
% PANDA -- Use the Panda algorithm to decompose A
%
% Usage:
%            [B, C] = PANDA(A, k)       returns rank-k Boolean decomposition of A
%      [B, C, time] = PANDA(A, k)       returns also the time
% [B, C, time, err] = PANDA(A, k)       returns also the error
%            [B, C] = PANDA(A, k, r)    uses r random iterations (default: 0)
%            [B, C] = PANDA(A, k, r, v) uses v verbosity (default: 0)
%
% See also: ASSO

error(nargchk(2, 4, nargin));

if nargin < 3, r = 0; end
if nargin < 4, v = 0; end

% Get the full path to panda
ppath = [fileparts(mfilename('fullpath')) filesep 'panda' filesep];
if v > 1, fprintf('Path to panda is ''%s''\n', ppath); end

% Get some temp files
Afile = tempname();
outfile = tempname();
Bfile = tempname();
Cfile = tempname();

if v > 1, 
    fprintf('Temp files are:\n Afile = %s\n outfile = %s\n Bfile = %s\n Cfile = %s\n', ...
          Afile, outfile, Bfile, Cfile);
end

if v > 0, fprintf('Writing data to disk... '); end

% Write A to out
fid = fopen(Afile, 'w');
for i=1:size(A, 1)
    ids = find(A(i, :) > 0);
    ids = ids - 1;  % Start indices from 0
    fprintf(fid, '%u ', ids); % MATLAB automatically vectorizes this
    fprintf(fid, '\n');
end
fclose(fid);

if v > 0, fprintf('done!\n'); end

% use Panda

% Update LD_LIBRARY_PATH
setenv('LD_LIBRARY_PATH', ...
'/usr/lib/:/usr/lib/matlab-7.11/sys/os/glnxa64:/usr/lib/matlab-7.11/bin/glnxa64:/usr/lib/matlab-7.11/extern/lib/glnxa64:/usr/lib/matlab-7.11/runtime/glnxa64:/usr/lib/matlab-7.11/sys/java/jre/glnxa64/jre/lib/amd64/native_threads:/usr/lib/matlab-7.11/sys/java/jre/glnxa64/jre/lib/amd64/server:/usr/lib/matlab-7.11/sys/java/jre/glnxa64/jre/lib/amd64');

if v > 0, fprintf('Running panda... '); end
tic
[status result] = unix([ppath 'topk -k' int2str(k) ' -r' int2str(r) ' -d' Afile ' -o' outfile ' -c1']);
time = toc;
if v > 0, fprintf('done!\n'); end
if status ~= 0
    error(['Panda returned an error code ' int2str(status) ' with text ' result]);
end

if v > 1, fprintf('Panda''s output was:\n* * *\n%s\n* * *\n', result); end

% reformat output
if v > 0, fprintf('Running parse_output.py... '); end
[status result] = system([ppath 'parse_output.py ' outfile ' ' Bfile ' ' Cfile]);
if v > 0, fprintf('done!\n'); end
if status ~= 0
    error(['Outfile parser returned an error code ' int2str(status) ' with text ' result]);
end
if v > 1, fprintf('parse_output.py''s output was:\n* * *\n%s\n* * *\n', result); end

% Read B and C (as full matrices)
if v > 0, fprintf('Reading B and C... '); end
B = full(spconvert(importdata(Bfile)));
C = full(spconvert(importdata(Cfile)));
% add extra rows to B
[n m] = size(A);
if size(B, 1) < n, B(n, 1) = 0; end
if size(C, 2) < m, C(1, m) = 0; end

if v > 0, fprintf('done!\n'); end

% Do we need to report error
if nargout > 3
    err = bnorm(A, B, C);
end

% Clean up temp files
delete(Afile, outfile, Bfile, Cfile);

